/* refine -- checks that alignments in a given species don't overlap. */

#define VERSION 2

#include "refine.h"
#include "maf.h"

#define CAUTIOUS

#define WIDTH 10                /* Default bandwidth */
#define SUFFIX "sc"
#define TBACK_SIZE 50000000
#define BUF_SIZE 5000
#define COL_WIDTH 75

/* Each of the main units of sequence data is called a "piece".  P[0] is
*  simply the "reference" sequence (species 0).  Piece P[i], for 0 < i < nP,
*  is the range of the i-th alignment.
*/

struct piece_end {
	int is_beg, piece_index, pseudo_colno;
	align_ptr align_col;
};

static void print_fasta(align_ptr align, int nRow, char **argv, char star) {
	align_ptr a;
	int i;
	char c;
/*	int nCol;

	for (nCol = 0, a = align; a != NULL; a = a->next)
		++nCol;

	printf("%d %d\n", nRow, nCol);
*/
	for (i = 0; i < nRow; ++i) {
		printf(">%s\n", argv[i+1]);
		for (a = align; a != NULL; a = a->next)
			putchar((c = a->col[i]) == '*' ? star : c);
		putchar('\n');
	}
}

static void print_fasta2(align_ptr align, int nRow, char **argv, char star) {
	align_ptr a;
	int i, col;
	char c;

	for (i = 0; i < nRow; ++i) {
		printf(">%s\n", argv[i+1]);
		for (col = 0, a = align; a != NULL; a = a->next) {
			putchar((c = a->col[i]) == star ? '-' : c);
			if (++col == COL_WIDTH) {
				putchar('\n');
				col = 0;
			}
		}
		if (col != 0)
			putchar('\n');
	}
}

double mafScore(struct mafAli maf, ss_t ss, int *gop, int *gtype) {
	uchar ai, ar, bi, br;
	int i;
	double score;
	struct mafComp *c1, *c2;

	if (ss == NULL)
		fatal("mafScoreRange: scores not initialized");

	score = 0.0;
	for (i = 0; i < maf.textSize; ++i) {
		for (c1 = maf.components; c1 != NULL; c1 = c1->next) {
			br = c1->text[i];
			for (c2 = c1->next; c2 != NULL; c2 = c2->next) {
				bi = c2->text[i];
			   	score += SS(br,bi);
				if (i > 0) {
					ar = c1->text[i-1];
					ai = c2->text[i-1];
					score -= GAP(ar,ai,br,bi);
				}
		   	}
		}
	}
	return score;
}

static void print_mafs(align_ptr align, piece_ptr P, int nP, int nRow,
  ss_t ss, int *gop, int *gtype, char *reference) {
	struct piece_end *ends, tmp;
	int i, j, k, n, p, r, x, beg, end, nc,
	   ncol = order_align(align),
	  *pirow = ckalloc(nRow*sizeof(int)),
	  *start = ckalloc(nRow*sizeof(int));
	struct mafAli A;
	struct mafComp *C = ckalloc(nRow*sizeof(struct mafComp));
	align_ptr a, cur_a;

	A.next = NULL;
	A.components = &C[0];
	for (i = 0; i < nRow; ++i)
		C[i].text = ckalloc((ncol+1)*sizeof(char));

	ends = ckalloc(2*nP*sizeof(struct piece_end));
	for (i = 0; i < nP; ++i) {
		j = 2*i;
		ends[j].is_beg = 1;
		ends[j+1].is_beg = 0;
		ends[j].piece_index = ends[j+1].piece_index = i;
		ends[j].align_col = P[i].firstcol;
		/* if one piece stops in the same column where another
		*  (presumably not in the same row) starts, process the
		*   start before the stop
		*/
		ends[j].pseudo_colno = 2*(P[i].firstcol->colno);
		ends[j+1].align_col = P[i].lastcol;
		ends[j+1].pseudo_colno = 2*(P[i].lastcol->colno) + 1;
	}
	for (i = 0; i < 2*nP; ++i)
		for (j = i+1; j < 2*nP; ++j)
			if (ends[j].pseudo_colno < ends[i].pseudo_colno) {
				tmp = ends[j];
				ends[j] = ends[i];
				ends[i] = tmp;
			}

/*
	for (i = 0; i < 2*nP; ++i) {
		j = ends[i].piece_index;
		printf("%d: piece %d (row %d) %s\n", ends[i].pseudo_colno / 2,
		  j, P[j].row, (ends[i].is_beg ? "starts" : "ends"));
	}
*/
	cur_a = P[0].firstcol;

	for (i = 0; i < nRow; ++i)
		pirow[i] = start[i] = -1;
	for (i = 0; i < 2*nP; i = j) {
		for (j = i+1; j < 2*nP; ++j)
			if (ends[j].pseudo_colno > ends[i].pseudo_colno)
				break;
		beg = ends[i].pseudo_colno/2 + (ends[i].is_beg == 0);
		if (j == 2*nP)
			end = ncol;
		else
			end = ends[j].pseudo_colno/2 - ends[j].is_beg;
		for (k = i; k < j; ++k) {
			x = ends[k].piece_index;
			r = P[x].row;
			if (ends[k].is_beg) {
				pirow[r] = x;
				start[r] = P[x].beg;
			} else
				start[r] = pirow[r] = -1;
		}
		if (beg > end)
			continue;
		A.textSize = end - beg + 1;
		a = cur_a;
		for (nc = k = 0; k < nRow; ++k)
			if ((p = pirow[k]) >= 0) {
				C[nc].src = P[p].name;
				C[nc].srcSize = P[p].srcSize;
				C[nc].start = start[k] - 1; // maf uses base 0
				C[nc].strand = P[p].strand; 
				C[nc].next = NULL;
				for (a = cur_a, n = 0;
				  n < A.textSize; ++n, a = a->next)
					if ((C[nc].text[n] =
					    a->col[k]) != '-')
						++start[k];
				if ((C[nc].size = start[k] - C[nc].start - 1)
				   == 0)
					continue;
				C[nc].text[n] = '\0';
				if (nc > 0)
					C[nc-1].next = &C[nc];
				++nc;
			}
		if (nc > 1 && same_string(A.components->src, reference)) {
			A.score = mafScore(A, ss, gop, gtype);
			mafWrite(stdout, &A);
		}
		cur_a = a;
	}
}


/* ---------------- Read the pieces and pairwise alignments --------------- */

static int count_alignments(char *reference, char *species, int fasta)
{
	FILE *fp;
	char align_file[BUF_SIZE], buf[BUF_SIZE];
	int count;

	sprintf(align_file, "%s.%s.%s", reference, species, SUFFIX);
	fp = ckopen(align_file, "r");
	if (fgets(buf, BUF_SIZE, fp) == NULL ||
	     !same_string(buf, "#:lav\n"))
		fatalf("%s is not a blastz output file", align_file);
	if (!fasta) {
		while (fgets(buf, BUF_SIZE, fp) != NULL &&
	     	!same_string(buf, "#:lav\n"))
			if (buf[0] == ' ')
				printf("# %s", buf);
			else
				printf("#\n");
	}
	
	count = 0;
	while (fgets(buf, BUF_SIZE, fp))
		if (strncmp(buf, "a {", 3) == 0)
			++count;
	fclose(fp);
	if (count == 0)
		fatalf("%s contains no alignments", align_file);
	return count;
}

static align_ptr new_align_node(int k)
{
	align_ptr new;
	int i;

	if (k <= 0)
		fatal("requesting empty alignment column");
	new = ckalloc(sizeof(align_node)); 
	new->col = ckalloc(k);
	for (i = 0; i < k; ++i)
		new->col[i] = FREE;
	new->colno = -1;
	return new;
}

static void piece_info(piece_ptr p, uchar *seq, char *name, int len, int beg,
  int end, int row, int srcSize, char strand)
{
	p->seq = ckalloc(len+1);
	strncpy((char *)p->seq, (char *)seq, len);
	p->name = copy_string(name);
	p->len = len;
	p->beg = beg;
	p->end = end;
	p->row = row;
	p->srcSize = srcSize;
	p->strand = strand;
}

/* generate a multiple alignment containing only the first sequence */
static align_ptr init_align(char *seqname, int nRow, piece_ptr P)
{
	SEQ *sf;
	align_ptr start = NULL, end = NULL, new;
	uchar *s;
	int i;

	sf = seq_open(seqname);
	if (!seq_read(sf) || SEQ_LEN(sf) == 0)
		fatal("First sequence cannot be read.");
	piece_info(P, SEQ_CHARS(sf), seqname, SEQ_LEN(sf), 1, SEQ_LEN(sf), 0,
	  SEQ_LEN(sf), '+');
	if (seq_read(sf))
		fatal("First sequence contains more than one contig.");
	seq_close(sf);

	s = P[0].seq;
	for (i = 0; i < P[0].len; i++) {
		new  = new_align_node(nRow);
		new->col[0] = (char)s[i];
		new->colno = i+1;
		if (start == NULL)
			start = new;
		new->prev = end;
		if (end != NULL)
			end->next = new;
		end = new;
	}
	end->next = NULL;
	start->prev = end;
	P[0].firstcol = start;
	P[0].lastcol = end;
	return start;
}

static void store(align_ptr ap, int pos, char c, char *sname)
{
	if (ap->col[pos] != FREE) 
		fatalf("Overlapping matches in %s to pos %d in base seq.", 
			sname, ap->colno);
	ap->col[pos] = c;
}

static char *ckget(char *buf, int size, FILE *fp, char *filename)
{
	char *ret;

	if ((ret = fgets(buf, size, fp)) == NULL)
		fatalf("error reading %s", filename);
	return ret;
}

static int pairwise(align_ptr align, char *reference, char *seq_name,
  int Pwhich, piece_ptr P, int nP, int nRow, int *row_ptr)
{
	piece_ptr p = &P[Pwhich];
	align_ptr new, ap = align;
	SEQ *sf = seq_open(seq_name);
	FILE *fp;
	uchar *contig = NULL;
	char buf[BUF_SIZE], buf2[BUF_SIZE], fname[BUF_SIZE],
	  *contig_name = NULL, *s, *t;
	int row = *row_ptr;
	int i, j, ncontigs, b1, b2, e1, e2, beg, end, len, prev_e2 = 0,
	  rc_curr, rc_lav, contig_curr, contig_lav;

	/* Throughout, ap->colno is the position (>= 1) in seq 0, else -1 */

	/* How many contigs in the sequence file? */
	fp = ckopen(seq_name, "r");
	ncontigs = 0;
	while (fgets(buf, BUF_SIZE, fp))
		if (buf[0] == '>')
			++ncontigs;
	fclose(fp);
	
	sprintf(fname, "%s.%s.%s", reference, seq_name, SUFFIX);
	fp = ckopen(fname, "r");
	ckget(buf, BUF_SIZE, fp, fname);
	if (!same_string(buf, "#:lav\n"))
		fatalf("%s is not a blastz output file.", buf2);

	rc_curr = contig_curr = 0;
  /* Read through the alignment. */
	while (fgets(buf, BUF_SIZE, fp))
    /* 1. LAV block */ if (same_string(buf, "#:lav\n")) {
		/* Get the new contig. */
		// Notice when we start on a new sequence, 
		// modulo reverse compliment, and read it
		while (ckget(buf, BUF_SIZE, fp, fname) && strcmp(buf, "s {\n"))
			;
		ckget(buf, BUF_SIZE, fp, fname);
		ckget(buf, BUF_SIZE, fp, fname);
		if (3 != sscanf(buf, " %s %*d %*d %d %d", buf2,
		  &rc_lav, &contig_lav))
			fatalf("bad sequence record {%s}", buf);
		if (contig_lav < contig_curr)
			fatal("impossibly small contig number");
		while (contig_curr < contig_lav) {
			if (!seq_read(sf))
				fatalf("Not enough contigs in %s.", seq_name);
			++contig_curr;
			rc_curr = 0;
		}
		if (rc_curr != rc_lav) {
			seq_revcomp_inplace(sf);
			rc_curr = rc_lav;
		}
		if ((s = strstr(buf2, "-\"")) != 0) {// XXX -- arggg
			*s++ = '"';
			*s++ = 0;
		}

		// find headers
		while (ckget(buf, BUF_SIZE, fp, fname) && strcmp(buf, "h {\n"))
			;
		ckget(buf, BUF_SIZE, fp, fname);
		ckget(buf, BUF_SIZE, fp, fname);
		if (strncmp(buf, "   \">", 5))
			fatalf("not a contig name: %s", buf);
		if (contig_name)
			free(contig_name);
		for (s = buf+5; *s == ' '; ++s)
			; 
		for (t = s; *t && *t != ' ' && *t != '"'; ++t)
			;
		*t = '\0';
		if (rc_curr && t[-1] == '-')
			t[-1] = '\0';
		contig_name = copy_string(s);
		contig = SEQ_CHARS(sf);
		ap = align;
    /* 2. Alignment */ } else if (same_string(buf, "a {\n")) {
		/* Record the new piece. */
		if (Pwhich >= nP)
			fatal("Too many pieces.");
		p = &P[Pwhich++];
		ckget(buf, BUF_SIZE, fp, fname);
		ckget(buf, BUF_SIZE, fp, fname);
		if (sscanf(buf, "  b %*d %d", &beg) != 1)
			fatalf("Cannot find beginning: %s", buf);
		ckget(buf, BUF_SIZE, fp, fname);
		if (sscanf(buf, "  e %*d %d", &end) != 1)
			fatalf("Cannot find end: %s", buf);
		len = end - beg + 1;
		strcpy(buf2, seq_name);
		if (ncontigs > 1) {
			strcat(buf2, ":");
			strcat(buf2, contig_name);
		}
		for (s = buf2; *s; ++s)
			if (strchr(" \t\n", *s))
				break;
		*s = '\0';
		piece_info(p, contig+beg-1, buf2, len, beg, end, row,
		  SEQ_LEN(sf), (rc_lav == 0 ? '+' : '-'));
		prev_e2 = 0;
    /* 3. Gap-free aligned segment */	} else if
		(sscanf(buf," l %d %d %d %d",&b1,&b2,&e1,&e2) == 4) {
	/* 3a. Handle gap between segments with extra characters in the current
	    piece. */
		if (prev_e2) {
		    if (ap->colno != b1-1 && prev_e2 != b2-1)
			fatal("bad inter-segment spacing");
		    for (j = prev_e2+1; j < b2; ++j) {
			/* For each extra letter in the current piece either: */
			if ((ap->next)->col[0] == DASH) {
				/* Add entry to existing column. */
				ap = ap->next;
				store(ap, row, contig[j-1], seq_name);
			} else {
				/* Create new column. */
				new = new_align_node(nRow);
				for (i = 0; i < row; ++i)
					new->col[i] = DASH;
				store(new, row, contig[j-1], seq_name);
				new->prev = ap;
				new->next = ap->next;
				new->next->prev = new;
				ap = ap->next = new;
			}
		    }
		}
	/* 3b. Move along links to the segment's start in the P[0].seq. */
		while (ap->colno < b1) {
			if ((ap = ap->next) == NULL)
				fatalf("crash 1 pos %d", ap->colno);
			if (prev_e2 && ap->colno < b1)
				/* Extra letter in P[0].seq; put a dash in
				   the current row. */
				store(ap, row, DASH, seq_name);
		}
	/* 3c. Match positions in the gap-free segment. */
		if (prev_e2 == 0)
			p->firstcol = ap;
		j = b2 - 1;
		store(ap, row, contig[j++], seq_name);
		while (ap->colno < e1) {
			if ((ap = ap->next) == NULL)
				fatalf("crash 2 at %d", ap->colno);
			store(ap, row,
			  (ap->col[0] != DASH) ? contig[j++] : DASH,
			  seq_name);
		}
		p->lastcol = ap;
		prev_e2 = e2;
	}

	if (contig_name)
		free(contig_name);
	contig_name = NULL;
	seq_close(sf);
	fclose(fp);
	*row_ptr = row + 1;
	return Pwhich;
}
static void free_all(align_ptr a, piece_ptr P, int nP)
{
	align_ptr b;
	int i;

	for ( ; a; a = b) {
		b = a->next;
		free(a->col);
		free(a);
	}
	for (i = 0; i < nP; ++i) {
		free(P[i].seq);
		free(P[i].name);
	}
	free(P);
}

#ifdef CAUTIOUS
static int bad_pair(align_ptr a, int nRow)
{
	align_ptr b;
	int r;
	uchar x, y;

	for ( ; (b = a->next) != NULL; a = b) 
		for (r = 0; r < nRow; ++r) {
			x = a->col[r];
			y = b->col[r];
			if ((x == DASH && y == FREE) ||
			    (x == FREE && y == DASH) ) {
				fprintf(stderr,
				  "x = %c, y = %c, col %d, row %d\n",
				   x, y, a->colno, r); 
				return 1;
			}
		}
	return 0;
}

static int bad_free(align_ptr A, piece_ptr P, int nP)
{
	align_ptr a;
	int p, r, cn, c;

	for (cn = 1, a = A; a; ++cn, a = a->next)
		a->colno = cn;
	for (p = 0; p < nP; ++p) {
		r = P[p].row;
		for (a = P[p].firstcol; a != P[p].lastcol->next; a = a->next)
			if ((c = a->col[r]) != DASH && !isalpha(c)) {
				fprintf(stderr, 
				  "Unwanted character `%c' in alignment,", c);
				fprintf(stderr,
				  "  piece %d, row %d, column %d\n",
				  p, r, a->colno);
				return 1;
			}
	}
	return 0;
}

/* bad_nonfree checks that only FREE occurs between pieces in the alignment;
*   the alignment's entries are distroyed in the process */
static int bad_nonfree(align_ptr A, int nRow, piece_ptr P, int nP)
{
	align_ptr a;
	int p, r, cn;
	uchar *acol;

	for (cn = 1, a = A; a; ++cn, a = a->next)
		a->colno = cn;
	for (p = 0; p < nP; ++p) {
		r = P[p].row;
		for (a = P[p].firstcol; a != P[p].lastcol->next; a = a->next)
			a->col[r] = FREE;
	}
	for (a = A; a; a = a->next) {
		acol = a->col;
		for (r = 0; r < nRow; ++r)
			if (acol[r] != FREE) {
				fprintf(stderr,
				  "`%c' between pieces at row %d, col %d\n",
				  acol[r], r, a->colno);
				return 1;
			}
	}
	return 0;
}
#endif

#define DEFAULT_ITER 1000
#define DEFAULT_DEPTH 10
int main(int argc, char *argv[])
{
	align_ptr align, ap;
	char star = '-', *scorefile, **x, cmd[100];
	gap_scores_t ds;
	ss_t ss;
	int gop[256], gtype[128];
	int Pwhich, nRow, row, score, i, nS, width, nP, new_score, flag_iter,
	  flag_depth, r, p, fasta = 0;
	piece_ptr P;
	TBack tback;

	sprintf(cmd, "refine.v%d", VERSION);
	argv0 = cmd;
	if (argc < 2)
    fatal("seq1 seq2 .. seqN [w=<width> [Q=scores] [TD=?] [fasta] [fasta2]");

	if (strncmp(argv[argc-1], "fasta2", 6) == 0) {
		if (strlen(argv[argc-1]) > 6)
			star = argv[argc-1][6];
		fasta = 2;
		--argc;
	}
	if (strncmp(argv[argc-1], "fasta", 5) == 0) {
		if (strlen(argv[argc-1]) > 5)
			star = argv[argc-1][5];
		fasta = 1;
		--argc;
	}
	/* Determine number of species. */
	for (nS = argc-1; strchr(argv[nS], '='); --nS) 
		;

	ckargs("QwTD", argc, argv, nS+1);
	get_argval_pos('w', &width, WIDTH);
	get_argval_nonneg('T', &flag_iter, DEFAULT_ITER);
	get_argval_nonneg('D', &flag_depth, DEFAULT_DEPTH);
	nRow = nS;

	if (!fasta) {
		mafWriteStart(stdout, argv0);
		printf("# %s", argv0);
		for (i = 0, x = argv; i < argc - 1; ++i)
			printf(" %s", *(++x));
		putchar('\n');
	}

	for (nP = 1, i = 2; i <= nS; ++i)
		nP += count_alignments(argv[1], argv[i], fasta);

	/* Set alignment scores. */
	if (get_cargval('Q', &scorefile))
		scores_from_file(scorefile, ss);
	else
		DNA_scores(ss);
	ds.E = 30;
	ds.O = 400;
	for (i = 0; i < 128; ++i)
		ss[i][DASH] = ss[DASH][i] = -ds.E;
	ss[DASH][DASH] = 0;
	for (i = 0; i < 128; ++i)
		ss[i][FREE] = ss[FREE][i] = 0;
	gap_costs(gop, gtype, ds.O);

	P = (piece_ptr)ckalloc(nP*sizeof(piece));

	/* Create an alignment containing only sequence 0. */
	align = init_align(argv[1], nRow, P);
	
	/* Build the crude multiple alignment from the pairwise alignments,
	*  determine the Pieces.
	*/
	for (row = 1, Pwhich = i = 1; i < nS; ++i)
		Pwhich = pairwise(align, argv[1], argv[i+1], Pwhich, P, nP,
		  nRow, &row);
	nRow = row;
#ifdef CAUTIOUS
	if (bad_free(align, P, nP))
		fatal("Bad FREE in very initial alignment.");
#endif

	for (ap = align; ap; ap = ap->next)
		for (r = 0; r < nRow; ++r)
			if (ap->col[r] == DASH)
				ap->col[r] = FREE;
	for (p = 0; p < nP; ++p) {
		r = P[p].row;
		for (ap = P[p].firstcol; ap!=P[p].lastcol->next; ap = ap->next)
			if (ap->col[r] == FREE)
				ap->col[r] = DASH;
		}

#ifdef CAUTIOUS
	if (bad_free(align, P, nP))
		fatal("Bad FREE in initial alignment.");
	if (bad_pair(align, nRow))
		fatal("Bad FREE/DASH pair in initial alignment");
#endif
	score = align_score(align, align->prev, nRow, ss, gop, gtype);


	if (flag_iter) {
		tback = ckalloc(sizeof(TB));
		tback->space = ckalloc(TBACK_SIZE*sizeof(uchar));
		tback->len = TBACK_SIZE;
		align = robin(align, nRow, P, nP, width, ss, gop, gtype,
		   flag_iter, flag_depth, tback);
/*
		new_score = align_score(align, align->prev, nRow, ss, gop,
		  gtype);
		fprintf(stderr, "Score changed from %d to %d.\n",
		  score, new_score);
*/
		score = new_score;
		free(tback->space);
		free(tback);
	}

	if (fasta == 1)
		print_fasta(align, nRow, argv, star);
	else if (fasta == 2)
		print_fasta2(align, nRow, argv, star);
	else
		print_mafs(align, P, nP, nRow, ss, gop, gtype, argv[1]);

#ifdef CAUTIOUS
	if (bad_pair(align, nRow))
		fatal("Bad FREE/DASH pair in final alignment");
	if (bad_free(align, P, nP))
		fatal("Bad FREE in final alignment.");
	if (bad_nonfree(align, nRow, P, nP))
		fatal("Bad non-FREE in final alignment.");
#endif

	free_all(align, P, nP);


	return 0;
}
